
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include "defines.h"
#include "model.h"
#include "minilzo.h"

#define _DPRINTF 1
#define DPRINTF if (_DPRINTF) printf

extern int console;

typedef unsigned char byte;

char *format_number(int number)
{
	static int idx=0;
	static char buffer[8][16];
	int cnt2, cnt3;

	idx=(idx+1)%8;

	sprintf(buffer[idx], "%d", number);
	for (cnt2=strlen(buffer[idx])-3; cnt2>0; cnt2-=3)
	{
		for (cnt3=strlen(buffer[idx])+1; cnt3>cnt2; cnt3--)
		{
			buffer[idx][cnt3]=buffer[idx][cnt3-1];
		}
		buffer[idx][cnt2]=',';
	}

	return buffer[idx];
}

void InitRawModel(model_t *mdl)
{
	memset(mdl, 0, sizeof(model_t));
	mdl->mesh=NULL;
	mdl->nmeshes=0;
	mdl->nfaces=0;
	mdl->vpkt.buffer=NULL;
	mdl->vpkt.buffer_size=0;
	mdl->type=MODEL_BAD_FORMAT;
	//mdl->obj_rot0[0]=3.1415926f/2.0f;
}

void FreeRawModel(model_t *mdl)
{
	int i;
	if (mdl->vpkt.buffer!=NULL) free(mdl->vpkt.buffer);
	for (i=mdl->nmeshes-1; i>=0; i--)
	{
		free(mdl->mesh[i].st);
		free(mdl->mesh[i].normal);
		free(mdl->mesh[i].vertex);
	}
	free(mdl->mesh);
	// DPRINTF("Free model done\n");
	InitRawModel(mdl);
}

void LoadFileToRawModel(char *filename, model_t *mdl)
{
	DPRINTF("Loading \"%s\"\n", filename);
	if (strstr(filename, "stripped")!=NULL)
		LoadStrippedToRawModel(filename, mdl);
	else
		mdl->type=MODEL_BAD_FORMAT;
}

void LoadStrippedToRawModel(char *filename, model_t *mdl)
{
	int filesize;
	char tmp_path[256];
	FILE *sf;
	unsigned int *indexes=NULL;

	strcpy(tmp_path, FILEHOST);
	FreeRawModel(mdl);

	sf=fopen(filename, "rb");
	if (sf==NULL)
	{
		DPRINTF("Could not load file \"%s\"\n", filename);
		mdl->type=MODEL_BAD_FORMAT;
		return;
	}

//	printf("r 1\n");

	fseek(sf, 0, SEEK_END);
	filesize=ftell(sf);
	fseek(sf, 0, SEEK_SET);

	DPRINTF("filesize: %s bytes\n", format_number(filesize));
//	printf("r 2\n");

	if (filesize<13)
	{
		DPRINTF("Unexpected end of stripped file\n");
		fclose(sf);
		mdl->type=MODEL_BAD_FORMAT;
	}
	else
	{
		unsigned int i;
		struct {
			char magic[8];
			unsigned char version;
			unsigned int nmeshes;
			unsigned int ntextures;
		} header;

		fread(&header, 1, 9, sf);

		if (strncmp(header.magic, STRIPPED_MAGIC, strlen(STRIPPED_MAGIC))!=0)
		{
			DPRINTF("Wrong file magic, prolly not a stripped file\n");
			mdl->type=MODEL_BAD_FORMAT;
			return;
		}

		if (header.version!=STRIPPED_VERSION)
		{
			DPRINTF("Wrong stripped file version\n");
			mdl->type=MODEL_BAD_FORMAT;
			return;
		}

		fread(&header.nmeshes, 4, 1, sf);
		fread(&header.ntextures, 4, 1, sf);
		indexes=(unsigned int *)malloc(header.ntextures*sizeof(unsigned int));

		for (i=0; i<header.ntextures; i++)
		{
			unsigned char len;
			fread(&len, 1, 1, sf);
			*(tmp_path+strlen(FILEHOST)+len)=0;
			fread(tmp_path+strlen(FILEHOST), 1, len, sf);
			// indexes[i]=AddTexture(tmp_path);
		}

		mdl->nfaces=0;
		mdl->nmeshes=header.nmeshes;

		mdl->mesh=(mesh_t*)memalign(64, mdl->nmeshes*sizeof(mesh_t));
		memset(mdl->mesh, 0, mdl->nmeshes*sizeof(mesh_t));

		for (i=0; i<mdl->nmeshes; i++)
		{
			fread(&mdl->mesh[i].texture, sizeof(unsigned int), 1, sf);
			//printf("texture: %d\n", mdl->mesh[i].texture);

			if (mdl->mesh[i].texture>=0)
			{
				mdl->mesh[i].texture=indexes[mdl->mesh[i].texture];
			}

			fread(&mdl->mesh[i].rgb, sizeof(float), 4, sf);
			fread(&mdl->mesh[i].nfaces, sizeof(unsigned int), 1, sf);

			{
				lzo_uint in_len, out_len1, out_len2, overhead;
				lzo_byte *in, *out, *wrkmem;
				out_len1=mdl->mesh[i].nfaces*VERT_DATA_LEN_STRIP;
				overhead=out_len1/64+16+3;
				mdl->mesh[i].vertex=(char *)memalign(64, out_len1);
				out=(lzo_byte *)mdl->mesh[i].vertex;
				// printf("v0 %d\n", *(((unsigned int *)out)-1));
				fread(&in_len, sizeof(lzo_uint), 1, sf);
				in=(lzo_byte *)memalign(64, in_len);
				wrkmem=(lzo_byte *)memalign(64, overhead);
				fread(in, 1, in_len, sf);
				lzo1x_decompress(in, in_len, out, &out_len2, NULL);
				free(wrkmem);
				free(in);
				if(out_len1!=out_len2)
				{
					DPRINTF("Error decompressing stripped file\n");
					FreeRawModel(mdl);
					return;
				}
			}
			{
				lzo_uint in_len, out_len1, out_len2, overhead;
				lzo_byte *in, *out, *wrkmem;
				out_len1=mdl->mesh[i].nfaces*NORM_DATA_LEN_STRIP;
				overhead=out_len1/64+16+3;
				mdl->mesh[i].normal=(char *)memalign(64, out_len1);
				out=(lzo_byte *)mdl->mesh[i].normal;
				// printf("n0 %d\n", *(((unsigned int *)out)-1));
				fread(&in_len, sizeof(lzo_uint), 1, sf);
				in=(lzo_byte *)memalign(64, in_len);
				wrkmem=(lzo_byte *)memalign(64, overhead);
				fread(in, 1, in_len, sf);
				lzo1x_decompress(in, in_len, out, &out_len2, NULL);
				free(wrkmem);
				free(in);
				if(out_len1!=out_len2)
				{
					DPRINTF("Error decompressing stripped file\n");
					FreeRawModel(mdl);
					return;
				}
			}
			
			if (mdl->mesh[i].texture>=0)
			{
				lzo_uint in_len, out_len1, out_len2, overhead;
				lzo_byte *in, *out, *wrkmem;
				out_len1=mdl->mesh[i].nfaces*ST_DATA_LEN_STRIP;
				overhead=out_len1/64+16+3;
				mdl->mesh[i].st=(char *)memalign(64, out_len1);
				out=(lzo_byte *)mdl->mesh[i].st;
				// printf("s0 %d\n", *(((unsigned int *)out)-1));
				fread(&in_len, sizeof(lzo_uint), 1, sf);
				in=(lzo_byte *)memalign(64, in_len);
				wrkmem=(lzo_byte *)memalign(64, overhead);
				fread(in, 1, in_len, sf);
				lzo1x_decompress(in, in_len, out, &out_len2, NULL);
				free(wrkmem);
				free(in);
				if(out_len1!=out_len2)
				{
					DPRINTF("Error decompressing stripped file\n");
					FreeRawModel(mdl);
					return;
				}
			}

			// we actually read the number of verts, correct that
			mdl->mesh[i].nfaces-=2;
			mdl->nfaces+=mdl->mesh[i].nfaces;
		}

		fread(&header.magic, 1, 8, sf);
		if (strncmp(header.magic, "STRIPPED", 8)!=0)
		{
			DPRINTF("Unexpected string at the end of stripped file.\n");
			FreeRawModel(mdl);
			mdl->type=MODEL_BAD_FORMAT;
			return;
		}

		fclose(sf);

		free(indexes);
	}
	
	DPRINTF("Loading done, %s meshes, %s polys\n", format_number(mdl->nmeshes), format_number(mdl->nfaces));

	mdl->type=MODEL_STRIPPED;
}


void GetFileList(const char*filename, filelist_t *list)
{
	char *ptr, *start_line, *end_line, *rot_ptr;
	void *data=NULL;
	int size;
	float *rot;
	ps2_vu0_fvector *rots;

	memset(list, 0, sizeof(filelist_t));

	{
		FILE *file;
		file=fopen(filename, "rb");
		if (file==NULL)
		{
			printf("Could not open file \"%s\" for reading (file: \"%s\", line: %d)\n", filename, __FILE__, __LINE__);
			return;
		}
		fseek(file, 0, SEEK_END);
		size=ftell(file);
		fseek(file, 0, SEEK_SET);
		printf("opened succesfully, size: %d\n", size);
		data=memalign(64, size);
		fread(data, 1, size, file);
		fclose(file);
	}

	ptr=data;
	while ( (((int)ptr-(int)data)<size-1) && (ptr[0]!=0) )
	{
		start_line=ptr;
		while (isspace(start_line[0])) start_line++;
		if (start_line[0]=='#')
		{
			while (start_line[0]!=0 && start_line[0]!='\n') start_line++;
			ptr=start_line+(start_line[0]!=0);
		}
		else if (start_line[0]!=0 && start_line[0]!='\n')
		{
			end_line=start_line;
			while (end_line[0]!=0 && end_line[0]!='\n') end_line++;
			end_line[0]=0;
			ptr=end_line+1;

			list->name_count++;
			if ( (list->name_count*sizeof(char*)) > list->allocated)
			{
				list->allocated+=1024;
				list->names=(char**)realloc((void*)list->names, list->allocated);
				list->rots=(ps2_vu0_fvector*)realloc((void*)list->rots, list->allocated*4);
			}

			memset(list->rots[list->name_count-1], 0, sizeof(ps2_vu0_fvector));

			rots=&list->rots[(list->name_count-1)];
			rot=(float*)rots;
			rot_ptr=start_line;

			for (; rot_ptr<end_line; rot_ptr++)
			{
				if (*rot_ptr==',')
				{
					*rot_ptr++=0;
					while (isspace(*rot_ptr)) rot_ptr++;
					*rot++=atof(rot_ptr);
				}
			}

			*rots[0]=*rots[0]*M_PI/180.0f;
			*rots[1]=*rots[0]*M_PI/180.0f;
			*rots[2]=*rots[0]*M_PI/180.0f;
			while (*rots[0]> M_PI) *rots[0]-=2.0f*M_PI;
			while (*rots[0]<-M_PI) *rots[0]+=2.0f*M_PI;
			while (*rots[1]> M_PI) *rots[1]-=2.0f*M_PI;
			while (*rots[1]<-M_PI) *rots[1]+=2.0f*M_PI;
			while (*rots[2]> M_PI) *rots[2]-=2.0f*M_PI;
			while (*rots[2]<-M_PI) *rots[2]+=2.0f*M_PI;
			*rots[3]=1.0f;

			list->names[list->name_count-1]=(char*)memalign(64, strlen(start_line)+1);
			strcpy(list->names[(list->name_count-1)],start_line);
		}
		else if (ptr[0]!=0) ptr++;
	}

	free(data);

	DPRINTF("Loading file list done (%d files)\n", list->name_count);
	DPRINTF("-\n");

	list->index=0;
}

