
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ps2gs.h>
#include <ps2vpu.h>
#include <malloc.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include "ps2hik.h"
#include "loaders.h"

typedef struct {
    int ch;
    int w, h;
    int dx, dy;
} display_t;

char *basename(char *filename);
void dirname(char *dest, char *filename);
int ps2hik_load_prefs(const char *filename);
int ps2hik_save_prefs(const char *filename);
void ps2hik_check_param(const char *p);
void ps2hik_save_display(display_t *display);
void ps2hik_restore_display(display_t *display);

char *program_name;
ps2_gs_gparam *gparam = NULL;
ps2_vpu *vpu0_handle=NULL;
ps2_vpu *vpu1_handle=NULL;
graph_settings_t graph_settings;
int fd_gs=-1;
void (*ps2hik_deinit)(int)=NULL;
int ps2hik_odev, ps2hik_frame=0;
display_t saved_display;

int ps2hik_release_devices(void)
{
	printf("release\n");
	if (fd_gs>-1) ps2_gs_close();
	if (vpu1_handle!=NULL) ps2_vpu_close(vpu1_handle);
	if (vpu0_handle!=NULL) ps2_vpu_close(vpu0_handle);

	fd_gs = -1;
	vpu1_handle = NULL;
	vpu0_handle = NULL;

	return PS2_GS_VC_ACQ_SUCCESS;
}

int ps2hik_acquire_devices(void)
{
	printf("acquire\n");
	if (fd_gs==-1) fd_gs = ps2_gs_open(-1);
	if (vpu0_handle==NULL) vpu0_handle = ps2_vpu_open(0);
	if (vpu1_handle==NULL) vpu1_handle = ps2_vpu_open(1);

	if ( (vpu0_handle==NULL) || (vpu1_handle==NULL) || (fd_gs==-1) )
	{
		printf("-- ERROR: could not open device\n");
		ps2hik_release_devices();
		return PS2_GS_VC_ACQ_FAILURE;
	}

	ps2_vpu_reset(vpu1_handle);
	ps2_vpu_reset(vpu0_handle);

	ps2_gs_reset(0, graph_settings.inter, graph_settings.out_mode, graph_settings.ff_mode, graph_settings.resolution, graph_settings.refresh_rate);

	return PS2_GS_VC_ACQ_SUCCESS;
}

int ps2hik_init(int argc, char **argv)
{
	char *pref_file;
	void (*sig_handle)(int)=SIG_ERR;
	char *program_dir;
	FILE *vu_file;

	graph_settings.inter = PS2_GS_NOINTERLACE;
	graph_settings.out_mode = PS2_GS_VESA;
	graph_settings.ff_mode = PS2_GS_FRAME;
	graph_settings.resolution = PS2_GS_640x480;
	graph_settings.refresh_rate = PS2_GS_60Hz;
	graph_settings.psm = PS2_GS_PSMCT32;
	graph_settings.zpsm = PS2_GS_PSMZ24;
	graph_settings.zbits = 24;
	graph_settings.cx = 0;
	graph_settings.cy = 0;

	program_name=basename(argv[0]);

	program_dir=(char*)malloc(strlen(argv[0])+1);
	dirname(program_dir, argv[0]);
	strcat(program_dir, "data/");
	chdir(program_dir);
	free(program_dir);

	gparam = ps2_gs_get_gparam();

	pref_file=(char*)malloc(strlen(basename(argv[0]))+6);
	sprintf(pref_file, "%s.pref", basename(argv[0]));

	ps2hik_load_prefs(pref_file);

	for (--argc; argc>0; --argc) ps2hik_check_param(argv[argc]);

	fd_gs = ps2_gs_open(-1);
	if (fd_gs==-1) { printf("Could not open " PS2_DEV_GS "\n"); return 0; }
	vpu0_handle = ps2_vpu_open(0);
	if (vpu0_handle==NULL) { printf("Could not open " PS2_DEV_VPU0 "\n"); return 0; }
	vpu1_handle = ps2_vpu_open(1);
	if (vpu1_handle==NULL) { printf("Could not open " PS2_DEV_VPU1 "\n"); return 0; }

	ps2_vpu_reset(vpu1_handle); // reset vpu1 first
	ps2_vpu_reset(vpu0_handle);

	vu_file=fopen("vu1render.vbin", "ro");
	if (vu_file==NULL)
	{
		printf("Could not open vu1render.vbin\n");
		return 0;
	}
	fread(ps2_vpu_text(vpu1_handle), 1, 16384, vu_file);
	fclose(vu_file);

	ps2_gs_vc_graphicsmode();

	ps2_gs_reset(0, graph_settings.inter, graph_settings.out_mode, graph_settings.ff_mode, graph_settings.resolution, graph_settings.refresh_rate);

	graph_settings.next2k=ps2_gs_set_dbuff(&graph_settings.db, graph_settings.psm, gparam->width, gparam->height, (graph_settings.zbits == 0) ? 0 : PS2_GS_ZGREATER, graph_settings.zpsm, 1);
	*(__u64 *)&graph_settings.db.clear0.rgbaq = PS2_GS_SETREG_RGBAQ(0x50, 0x10, 0x20, 0x80, 0x3f800000);
	*(__u64 *)&graph_settings.db.clear1.rgbaq = PS2_GS_SETREG_RGBAQ(0x50, 0x10, 0x20, 0x80, 0x3f800000);
	ps2_gs_put_drawenv(&graph_settings.db.giftag1);
	ps2_gs_set_finish(&graph_settings.finish);
	ps2_gs_wait_finish(&graph_settings.finish);

	graph_settings.tex_addr=graph_settings.next2k*2048/64;

	ps2_gs_sync_v(0);

	ps2_gs_start_display(1);
	ps2_gs_vc_enablevcswitch(ps2hik_acquire_devices, ps2hik_release_devices);

	graph_settings.width=gparam->width;
	graph_settings.height=gparam->height;

	sig_handle=signal(SIGINT, ps2hik_release);
	if (sig_handle==SIG_ERR) printf("Could not trap signal SIGINT\n");
	sig_handle=signal(SIGTERM, ps2hik_release);
	if (sig_handle==SIG_ERR) printf("Could not trap signal SIGTERM\n");

	ps2hik_save_prefs(pref_file);

	free(pref_file);

	return 1;
}

void ps2hik_release(int sig)
{
	if (sig) printf("Recieved signal %d, quitting.\n", sig);
	if (ps2hik_deinit!=NULL) ps2hik_deinit(sig);

	//ps2hik_release_devices();

	exit(sig);
}

void (*ps2hik_set_deinit(void (*handler)(int)))(int)
{
	void (*old_deinit)(int)=ps2hik_deinit;
	ps2hik_deinit=handler;
	return old_deinit;
}

int ps2hik_load_prefs(const char *filename)
{
	FILE *pref_file;
	char *pref_text, *start, *end;
	int pref_size;

	pref_file=fopen(filename, "ro");
	if (pref_file==NULL) return 0;

	fseek(pref_file, 0, SEEK_END);
	pref_size=ftell(pref_file);
	fseek(pref_file, 0, SEEK_SET);

	start=pref_text=(char*)malloc(pref_size+1);
	fread(pref_text, 1, pref_size, pref_file);
	start[pref_size]=0;

	fclose(pref_file);

	while ( (!isalnum(*start)) && (*start!=0) ) start++;

	while (*start!=0)
	{
		end=start+1;
		while (isalnum(*end)) end++;
		if (*end==0) end--; else *end=0;
		ps2hik_check_param(start);
		start=end+1;
	}

	free(pref_text);

	return 1;
}

void usage()
{
	printf("Usage: %s [OPTION] \n", program_name);
	printf("Options:\n"
 	"-inter: set interlace mode\n"
 	"-nointer: set non-interlace mode\n"
 	"-ntsc: set display to NTSC\n"
 	"-pal: set display to PAL\n"
 	"-dtv: set display to DTV\n"
 	"-vesa: set display to VESA\n"
 	"-vga: set display to VESA (alias to VESA)\n"
 	"-frame: set FRAME mode\n"
 	"-field: set FIELD mode\n"
 	"-640x480: set resolution to 640x480\n"
 	"-800x600: set resolution to 800x600\n"
 	"-1024x768: set resolution to 1024x768\n"
 	"-1280x1024: set resolution to 1280x1024\n"
 	"-60: set frequency to 60Hz\n"
 	"-75: set frequency to 75Hz\n"
 	"-32: set depth to 32bit\n"
 	"-24: set depth to 24bit\n"
 	"-16: set depth to 16bit\n"
 	"-z24: use a 24bit zbuffer\n"
 	"-z16: use a 16bit zbuffer\n"
 	"-z0:  disable the zbuffer\n"
	"-cxN: center the screen on X by N pixels\n"
	"-cyN: center the screen on Y by N pixels\n");
	printf("Options are saved in data/%s.pref\n", program_name);
}

int ps2hik_save_prefs(const char *filename)
{
	FILE *pref_file;

	pref_file=fopen(filename, "w");
	if (pref_file==NULL)
	{
		printf("Cannot save prefs to \"%s\"\n", filename);
		return 0;
	}
	
 	if (graph_settings.inter == PS2_GS_INTERLACE)      fprintf(pref_file, "inter ");
 	if (graph_settings.inter == PS2_GS_NOINTERLACE)    fprintf(pref_file, "nointer ");
 	if (graph_settings.out_mode == PS2_GS_NTSC)        fprintf(pref_file, "ntsc ");
 	if (graph_settings.out_mode == PS2_GS_PAL)         fprintf(pref_file, "pal ");
 	if (graph_settings.out_mode == PS2_GS_DTV)         fprintf(pref_file, "dtv ");
 	if (graph_settings.out_mode == PS2_GS_VESA)        fprintf(pref_file, "vesa ");
 	if (graph_settings.ff_mode == PS2_GS_FRAME)        fprintf(pref_file, "frame ");
 	if (graph_settings.ff_mode == PS2_GS_FIELD)        fprintf(pref_file, "field ");
 	if (graph_settings.resolution == PS2_GS_640x480)   fprintf(pref_file, "640x480 ");
 	if (graph_settings.resolution == PS2_GS_800x600)   fprintf(pref_file, "800x600 ");
 	if (graph_settings.resolution == PS2_GS_1024x768)  fprintf(pref_file, "1024x768 ");
 	if (graph_settings.resolution == PS2_GS_1280x1024) fprintf(pref_file, "1280x1024 ");
 	if (graph_settings.refresh_rate == PS2_GS_60Hz)    fprintf(pref_file, "60 ");
 	if (graph_settings.refresh_rate == PS2_GS_75Hz)    fprintf(pref_file, "75 ");
 	if (graph_settings.psm == PS2_GS_PSMCT32)          fprintf(pref_file, "32 ");
 	if (graph_settings.psm == PS2_GS_PSMCT24)          fprintf(pref_file, "24 ");
 	if (graph_settings.psm == PS2_GS_PSMCT16S)         fprintf(pref_file, "16 ");
	if (graph_settings.zbits == 24)                    fprintf(pref_file, "z24 ");
	if (graph_settings.zbits == 16)                    fprintf(pref_file, "z16 ");
	if (graph_settings.zbits ==  0)                    fprintf(pref_file, "z0 ");
	if (graph_settings.cx>0)                           fprintf(pref_file, "cx%d ", graph_settings.cx);
	if (graph_settings.cy>0)                           fprintf(pref_file, "cy%d ", graph_settings.cy);

	fclose(pref_file);

	return 1;
}

void ps2hik_check_param(const char *p)
{
	int everything_ok=1;

	if (*p=='-') p++;
	     if (!strcmp(p, "inter"))     { graph_settings.inter = PS2_GS_INTERLACE; }
	else if (!strcmp(p, "nointer"))   { graph_settings.inter = PS2_GS_NOINTERLACE; }
	else if (!strcmp(p, "ntsc"))      { graph_settings.out_mode = PS2_GS_NTSC; }
	else if (!strcmp(p, "pal"))       { graph_settings.out_mode = PS2_GS_PAL; }
	else if (!strcmp(p, "dtv"))       { graph_settings.out_mode = PS2_GS_DTV; }
	else if (!strcmp(p, "vesa"))      { graph_settings.out_mode = PS2_GS_VESA; }
	else if (!strcmp(p, "vga"))       { graph_settings.out_mode = PS2_GS_VESA; }
	else if (!strcmp(p, "frame"))     { graph_settings.ff_mode = PS2_GS_FRAME; }
	else if (!strcmp(p, "field"))     { graph_settings.ff_mode = PS2_GS_FIELD; }
	else if (!strcmp(p, "640x480"))   { graph_settings.resolution = PS2_GS_640x480; }
	else if (!strcmp(p, "800x600"))   { graph_settings.resolution = PS2_GS_800x600; }
	else if (!strcmp(p, "1024x768"))  { graph_settings.resolution = PS2_GS_1024x768; }
	else if (!strcmp(p, "1280x1024")) { graph_settings.resolution = PS2_GS_1280x1024; }
	else if (!strcmp(p, "60"))        { graph_settings.refresh_rate = PS2_GS_60Hz; }
	else if (!strcmp(p, "75"))        { graph_settings.refresh_rate = PS2_GS_75Hz; }
	else if (!strcmp(p, "32"))        { graph_settings.psm = PS2_GS_PSMCT32; }
	else if (!strcmp(p, "24"))        { graph_settings.psm = PS2_GS_PSMCT24; }
	else if (!strcmp(p, "16"))        { graph_settings.psm = PS2_GS_PSMCT16S; }
	else if (!strcmp(p, "z24"))       { graph_settings.zpsm = PS2_GS_PSMZ24; graph_settings.zbits = 24; }
	else if (!strcmp(p, "z16"))       { graph_settings.zpsm = PS2_GS_PSMZ16S; graph_settings.zbits = 16; }
	else if (!strcmp(p, "z0"))        { graph_settings.zpsm = PS2_GS_PSMZ24; graph_settings.zbits = 0; }
	else if (!strncmp(p, "cx", 2))    { graph_settings.cx =strtol(p+2, NULL, 10); }
	else if (!strncmp(p, "cy", 2))    { graph_settings.cy =strtol(p+2, NULL, 10); }
	else if (!strcmp(p, "help"))      { usage(); everything_ok=0; }
	else { printf("Invalid parameter: \"%s\", use -help.\n", p); everything_ok=0; }

	if (!everything_ok) exit(1);

	if (graph_settings.psm == PS2_GS_PSMCT16S && graph_settings.zpsm == PS2_GS_PSMZ16)
	{
		graph_settings.psm = PS2_GS_PSMCT16;
		graph_settings.zpsm = PS2_GS_PSMZ16;
	}
}

char *basename(char *filename)
{
	unsigned int s;
	char *ret=filename;
	for (s=0; s<strlen(filename)-1; s++)
	{
		if (filename[s]=='/') ret=filename+s+1;
	}
	return ret;
}

void dirname(char *dest, char *filename)
{
	char *ptr;
	strcpy(dest, filename);
	ptr=basename(dest);
	*ptr=0;
	//if (strlen(dest)==0) strcpy(dest, "./");
}

void ps2hik_save_display(display_t *display)
{
	int gs=-1;
	gs = open(PS2_DEV_GS, O_RDWR);

	if (gs < 0) { perror("Couldn't open " PS2_DEV_GS); return; }

	memset (display, 0, sizeof (struct ps2_display));
	display->ch = 0;

	if (ioctl(gs, PS2IOC_GDISPLAY, display) < 0)
		perror("Could not read DISPLAY1");

	close(gs);
}

void ps2hik_restore_display(display_t *display)
{
	int gs=-1;
	gs = open(PS2_DEV_GS, O_RDWR);

	if (gs < 0) { perror("Couldn't open " PS2_DEV_GS); return; }

	if (ioctl(gs, PS2IOC_SDISPLAY, display) < 0)
		perror("Could not set DISPLAY1");

	close(gs);
}

void ps2hik_lock()
{
	ps2_gs_vc_lock();

	ps2_gs_set_half_offset((ps2hik_frame & 1) ? &graph_settings.db.draw1 : &graph_settings.db.draw0, ps2hik_odev);
	ps2_gs_swap_dbuff(&graph_settings.db, ps2hik_frame);
}

void ps2hik_unlock()
{
	ps2hik_frame++;
	ps2hik_odev = !ps2_gs_sync_v(0);

	ps2_gs_vc_unlock();
}

void ps2hik_set_image_address(image_t *im)
{
	im->clut.fbp=graph_settings.tex_addr;
	graph_settings.tex_addr+=32;
	im->texture.fbp=graph_settings.tex_addr;
	switch (im->texture.psm)
	{
		case PS2_GS_PSMT8:
			graph_settings.tex_addr+=32*((im->texture.w<128)?1:(im->texture.w/128))*((im->texture.h<64)?1:(im->texture.h/64));
			break;
		default:
			printf("Couldn't get the texture size (file %s - line %d)\n", __FILE__, __LINE__);
			break;
	}
}

void ps2hik_printf_resolution()
{
	char *mode;

	switch (graph_settings.out_mode)
	{
		case PS2_GS_NTSC:
			mode="ntsc";
			break;
		case PS2_GS_PAL:
			mode="pal";
			break;
		case PS2_GS_DTV:
			mode="dtv";
			break;
		case PS2_GS_VESA:
			mode="vesa";
			break;
		default:
			mode="unknown";
			break;
	}

	printf("Using %dx%d (%s)\n", gparam->width, gparam->height, mode);
}





